(* WARHOL Abstract Syntax Tree *)

(* Operators *)
type op = Add | Subt | Mult | Div |
		Eq | Neq | Less | Leq | Greater | Geq |
		And | Or
type uop = Not | Neg

(*Types*)
type typ = Int | Void | Bool | String | Char
 | Matrix1DType of typ * int
 | Matrix1DPointer of typ

type bind = typ * string (*bind means the same thing as datatype in CMAT*)

(*Expressions*)
type expr =
  IntLiteral of int
| StringLiteral of string
| CharLiteral of char
| BoolLit of bool
| Id of string
| Noexpr
| Binop of expr * op * expr
| Unop of uop * expr
| Assign of expr * expr
| MatrixLiteral of expr list 	(*i.e. 1,1,1, I believe*)
| Matrix1DAccess of string * expr	(*i.e. i[2], I believe*)
| Matrix1DReference of string
| Dereference of string
| PointerIncrement of string
| Call of string * expr list

(*Statements*)
type stmt =
  Block of stmt list
| Expr of expr
| Return of expr
| If of expr * stmt * stmt
| While of expr * stmt
| For of expr * expr * expr * stmt

(*Function declarations*)
type func_body = {
    f_vdecls: bind list;
    f_stmts: stmt list;
  }

(*Function declarations*)
type func_decl = {
    return_type : typ;
    fname : string;
    formals : bind list;
    fbody : func_body;
}


(*Start symbol*)
type program = bind list * func_decl list


(* Pretty-printing functions *)

let string_of_op = function
    Add -> "+"
  | Subt -> "-"
  | Mult -> "*"
  | Div -> "/"
  | Eq -> "=="
  | Neq -> "!="
  | Less -> "<"
  | Leq -> "<="
  | Greater -> ">"
  | Geq -> ">="
  | And -> "&&"
  | Or -> "||"

let string_of_uop = function
    Neg -> "-"
  | Not -> "!"

let string_of_matrix m =
  let rec string_of_matrix_lit = function
      [] -> "]"
    | [hd] -> (match hd with
                IntLiteral(i) -> string_of_int i
              | _ -> raise( Failure("Illegal expression in matrix literal") )) ^ string_of_matrix_lit []
    | hd::tl -> (match hd with
                    IntLiteral(i) -> string_of_int i ^ ", "
                  | _ -> raise( Failure("Illegal expression in matrix literal") )) ^ string_of_matrix_lit tl
  in
  "[" ^ string_of_matrix_lit m

let rec string_of_expr = function
    IntLiteral(l) -> string_of_int l
  | CharLiteral(i) ->  String.make 1 i
  | StringLiteral(i) -> i
  | BoolLit(true) -> "true"
  | BoolLit(false) -> "false"
  | Id(s) -> s
  | Binop(e1, o, e2) ->
      string_of_expr e1 ^ " " ^ string_of_op o ^ " " ^ string_of_expr e2
  | PointerIncrement(s) -> "++" ^ s 
  | Unop(o, e) -> string_of_uop o ^ string_of_expr e
  | Assign(v, e) -> string_of_expr v ^ " = " ^ string_of_expr e
  | Call(f, el) ->
      f ^ "(" ^ String.concat ", " (List.map string_of_expr el) ^ ")"
  | Noexpr -> ""
  | MatrixLiteral(m) -> string_of_matrix m
  | Matrix1DAccess(s, r1) -> s ^ "[" ^ (string_of_expr r1) ^ "]"
  | Matrix1DReference(s) -> "@" ^ s
  | Dereference(s) -> "&" ^ s

let rec string_of_stmt = function
    Block(stmts) ->
      "{\n" ^ String.concat "" (List.map string_of_stmt stmts) ^ "}\n"
  | Expr(expr) -> string_of_expr expr ^ ";\n";
  | Return(expr) -> "return " ^ string_of_expr expr ^ ";\n";
  | If(e, s, Block([])) -> "if (" ^ string_of_expr e ^ ")\n" ^ string_of_stmt s
  | If(e, s1, s2) ->  "if (" ^ string_of_expr e ^ ")\n" ^
      string_of_stmt s1 ^ "else\n" ^ string_of_stmt s2
  | For(e1, e2, e3, s) ->
      "for (" ^ string_of_expr e1  ^ " ; " ^ string_of_expr e2 ^ " ; " ^
      string_of_expr e3  ^ ") " ^ string_of_stmt s
  | While(e, s) -> "while (" ^ string_of_expr e ^ ") " ^ string_of_stmt s

let rec string_of_typ = function
    Int -> "int"
  | Bool -> "bool"
  | Void -> "void"
  | String -> "string"
  | Matrix1DType(t, i1) -> string_of_typ t ^ "[" ^ string_of_int i1 ^ "]"
  | Matrix1DPointer(t) -> string_of_typ t ^ "[]"

let string_of_vdecl (t, id) = string_of_typ t ^ " " ^ id ^ ";\n"

let string_of_fdecl fdecl =
  string_of_typ fdecl.return_type ^ " " ^
  fdecl.fname ^ "(" ^ String.concat ", " (List.map snd fdecl.formals) ^
  ")\n{\n" ^
  String.concat "" (List.map string_of_vdecl fdecl.fbody.f_vdecls) ^
  String.concat "" (List.map string_of_stmt fdecl.fbody.f_stmts) ^
  "}\n"

let string_of_program (vars, funcs) =
  String.concat "" (List.map string_of_vdecl vars) ^ "\n" ^
  String.concat "\n" (List.map string_of_fdecl funcs)
